# ----------------------------------------------------------------------------
# Tcl Erweiterungs-Skript                                  Referenzenvariablen
# Copyright (c) 2002 by Hartmut Schwarz               matut@matut-spielwerk.de
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2 
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# ----------------------------------------------------------------------------

# Erweitert Tcl um Referenzen auf Variablen. Anstelle der eigentlichen Daten
# wird in einer Referenzvariablen lediglich eine Referenz auf diese Daten
# gespeichert. Das ist ein automatisch generierter globaler Name unter dem die
# Daten gespeichert sind. Damit dies geschehen kann, muss vorher genau einmal
# die Prozedur "REFinit" ausgefuehrt werden (geschieht bei Laden von "ref.tcl"
# von selbst). 

proc REFinit {} {
 foreach rname [info global REF*] {
  global $rname;
  unset $rname;
 }
 global REFcnt;
 set REFcnt -1;
}

# Beispiel: REFinit; # initialisiert die Referenzenvariablen-Erweiterung

# Referenzen koennen mit der Prozedur "REF" erzeugt werden. Die erzeugte 
# Referenz wird dabei unter dem angegebenen, damit als Referenzvariable
# deklarierten Namen, abgelegt.

proc REF name {
 upvar $name $name;
 global REFcnt;
 incr REFcnt;
 set $name REF$REFcnt;
 uplevel 1 "global [set $name]";
}

# Beispiel: REF n; # erzeugt neue Referenz unter der Referenzvariablen "n".

# Durch einfache Zuweisung der Referenz entstehen weitere Referenzvariablen
# die auf dieselben Daten verweisen:

# Beispiel: set 2 $n; # erzeugt Referenzvariable "n2" mit Verweis auf
#                     # dieselben Daten wie "n".
 
# Der Inhalt einer Referenzvariablen kann dann mit "REFset" gesetzt bzw.
# ausgegeben werden.

proc REFset { ref args } {
 upvar $ref r;
 global $r;
 set len [llength $args]; 
 if { $len == 1 } {
  set $r [lindex $args 0];
 } elseif { $len != 0 } {
  error "wrong # args: should be \"REFset varName ?newValue?\"";
 }
 return [set $r];
}

# Beispiel: 
# - den Inhalt der Referenzvariablen "n" auf "Paul" setzen:
# REFset n "Paul";
# - den Inhalt der Referenzvariablen "n" ausgeben. 
# REFset n; # liefert "Paul" 

# Eine einfache Substitution von Referenzvariablen mit "$" ist nicht moeglich.
# Diese liefert lediglich die Referenz, d.h. den globalen Referenznamen der
# Variablen.

# Der Vorteil von Referenzvariablen gegenueber normalen Variablen ist, dass
# man in komplexen Datenstrukturen von verschiedenen Stellen aus,
# insbesondere aus Feldern heraus, auf dieselben Daten verweisen kann.
# Aenderungen der Daten kommen so automatisch bei allen darauf verweisenden
# Referenzvariablen an. 

# Beispiel ohne Referenzvariable:
# set a "Peter";  # setzt "a" auf "Peter"
# set b $a;       # setzt "b" auf Inhalt von "a", also "Peter"
# set a "Paul";   # ueberschreibt Inhalt von "a" mit "Paul"
# set b;          # liefert immer noch "Peter"
# Beispiel mit Referenzvariablen:
# REF a;             # deklariert Referenzvariable "a"
# REFset a "Peter";  # setzt Referenzvariable "a" auf "Peter"
# set b $a;          # deklariert Referenzvariable "b" mit derselben Referenz
# REFset a "Paul";   # ueberschreibt Inhalt von "a" mit "Paul"
# REFset b;          # liefert den geaenderten Inhalt der Referenz "Paul".

# *** hier wird die Referenzvariablen-Erweiterung initialisiert ***
REFinit;
