# ----------------------------------------------------------------------------
# Csound Tcl Music-Script                             die Grundlagen der Musik
# Copyright (c) 2002 by Hartmut Schwarz               matut@matut-spielwerk.de
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2 
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# ----------------------------------------------------------------------------
set CP \
"; Copyright (c) 2002-2004 by Hartmut Schwarz, matut@matut-spielwerk.de
;
; This program is free software; you can redistribute it and/or
; modify it under the terms of the GNU General Public License
; as published by the Free Software Foundation; either version 2 
; of the License, or (at your option) any later version.
;
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
; GNU General Public License for more details.
;
; You should have received a copy of the GNU General Public
; License along with this program; if not, write to the Free
; Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
; MA 02111-1307, USA.
;
";
set VS "0.00.002";

# ----------------------------------------------------------------------------
puts "Grundwissen der Musik fuer den Tcl-Interpreter";
puts "Musical basics for the Tcl-Interpreter";
# ----------------------------------------------------------------------------

# ----------------------------------------------------------------------------
puts "Lektion x : Elemente des Spielkontextes CTX, play-context elements";
# ----------------------------------------------------------------------------

# 1. Wurzeleintraege, root entries
# CTX(loca)   = Ort (Ausgabedateiname)  = location (output file name)
# CTX(tttm)   = Gesamtzeit              = total time 
# CTX(out)    = Ausgabekanal            = output channel
# CTX(crin)   = aktuelles Instrument    = currently playing instrument
# CTX(instr)  = Instrumentenliste       = instrument list {inam1 inam2 ... }
# CTX(fstats) = Funktionstabellen       = f-Statements {fst1 fst2 ...}
# CTX(fnums)  = FT-Nummern              = ft-numbers {nm1 ftn1 nm2 ftn2 ... }

# 2. Stammeintraege, stem entries
# CTX(crtm)   = aktuelle Zeit (Sek.)    = current time in seconds
# CTX(sttm)   = Zeitstempel (Sek.)      = time stamp in seconds
# CTX(sttv)   = Zeitstempel (Notenwert) = time stamp as time value
# CTX(crtv)   = aktuelle Notenzeit      = current time as time value
# CTX(lega)   = Legato-Kennung          = legato indicator
# CTX(fort)   = piano/forte-Stufe       = piano/forte level
# CTX(velo)   = adagio/veloce-Stufe     = adagio/veloce level 
# CTX(iParN)  = Parameterspeicher       = parameter storage (N=1,2,..)
# CTX(opng)   = Eroeffnungsprozedur     = opening procedure
# CTX(prld)   = Vorspielprozedur        = prelude procedure
# CTX(song)   = Liedprozedur            = song procedure
# CTX(psld)   = Nachspielprozedur       = postlude procedure
# CTX(clng)   = Abschlussprozedur       = closing procedure
# CTX(bthk)   = Akkordschlagprozedur    = chord hook procedure
# CTX(inam)   = Instrumentenname        = instrument name
# CTX(tmsg)   = Taktangabe als "Z/N"    = time signature as "Z/N"
# CTX(tmtv)   = Zeit pro Notenwert      = time per time value 
# CTX(tvtv)   = Notendauer pro -wert    = note duration per time value 
# CTX(tmfn)   = Zeitfunktion            = time function
# CTX(kybs)   = Grundton                = key base
# CTX(kyfn)   = Tastennummerfunktion    = key code function
# CTX(btbs)   = Grundschlag             = beat base
# CTX(vlbs)   = Grundanschlagstaerke    = volume (=attac) base
# CTX(cres)   = Crescendo               = crescendo (volume per chord)
# CTX(fstp)   = piano/forte-Abstufung   = forte step
# CTX(vlfn)   = Anschlagstaerkefunktion = volume function
# CTX(loud)   = Lautstaerkefaktor       = loudness factor
# CTX(vstp)   = adagio/veloce-Abstufung = veloce step
# CTX(frmt)   = Fermate Verlaengerung   = fermata prolongation
# CTX(optv)   = Notenwert Eroeffnung    = opening time value
# CTX(prtv)   = Notenwert Vorspiel      = prelude time value
# CTX(pstv)   = Notenwert Nachspiel     = postlude time value 
# CTX(cltv)   = Notenwert Abschluss     = closing time value 
# CTX(pmod)   = Notenspielmodus         = play mode (melody or rhythm)
# CTX(csym)   = Akkordsymbol            = chord symbol
# CTX(chtv)   = Notenwert Akkord        = chord time value
# CTX(chbt)   = Akkord Schlaege         = chord beats
# CTX(cmod)   = Akkord-Modus            = chord mode "rs."
# CTX(cstc)   = Staccato-Reduktion      = staccato reduction
# CTX(text)   = Text                    = lyrics
# CTX(idst)   = Instrumentenabstand [%] = instrument distance [%]
# CTX(iazm)   = Instrumentenrichtung    = instrument direction 
# CTX(isze)   = Instrumentengroesse [%] = instrument size     [%]
# CTX(iang)   = Instrumentenstellung    = instrument orientation

# 3. Zweigeintraege, twig entries
# CTX(turn)   = naechstes an der Reihe  = next leaf in turn (-1=none)
# CTX(rept)   = Anzahl Wiederholungen   = number of repetitions 
# CTX(rlft)   = Wiederholungszaehler    = repetitions left counter
# CTX(vars)   = Varianten               = variations
# CTX(vind)   = Variantenindex          = variation index

# ----------------------------------------------------------------------------
puts "Lektion x : Allgemeine Grundlagen, common basics";
# ----------------------------------------------------------------------------

# Konstanten
set PI2      [expr 4.0*asin(1)];
set X12LN2   [expr 12.0/log(2.0)];
set CPS0     [expr 440.0*pow(2,-21.0/12.0)]; # Frequenz fuer Taste 0 (C)
set KBASE264 12.15641287;                    # Tastennummer fuer 264 Hz

# Registrierung statistischer Daten
set STAT_NumberOfHits  0;          # Gesamtzahl der Anschlaege
set STAT_DurationTotal 0;          # Gesamtsumme der Anschlagdauer
set STAT_DurationMax   0;          # maximale Anschlagdauer
set STAT_DurationMin   0xffffffff; # minimale Anschlagdauer
set STAT_VolumeTotal   0;          # Gesamtsumme Anschlagstaerke
set STAT_VolumeMax     0;          # maximale Anschlagstaerke
set STAT_VolumeMin     0xffffffff; # minimale Anschlagstaerke
set STAT_KeyTotal      0;          # Gesamtsumme Tastennummer
set STAT_KeyMax       -0xffffffff; # maximale Tastennummer
set STAT_KeyMin        0xffffffff; # minimale Tastennummer

proc register_statistics { rCTX dura vol key } {
# registriert Spielstatistik (z.Zt. nur Anschlagdauer)
 global $rCTX STAT_NumberOfHits\
        STAT_DurationTotal STAT_DurationMax STAT_DurationMin\
        STAT_VolumeTotal   STAT_VolumeMax   STAT_VolumeMin\
        STAT_KeyTotal      STAT_KeyMax      STAT_KeyMin;
 if { [string is double $key] } {
  set key [expr $key + [TREEget $rCTX kybs]];
  incr STAT_NumberOfHits;
  set STAT_DurationTotal [expr $STAT_DurationTotal + $dura];
  set STAT_VolumeTotal [expr $STAT_VolumeTotal + $vol];
  set STAT_KeyTotal [expr $STAT_KeyTotal + $key];
  if { $dura > $STAT_DurationMax } {
   set STAT_DurationMax $dura;
  }
  if { $dura < $STAT_DurationMin } {
   set STAT_DurationMin $dura;
  }
  if { $vol > $STAT_VolumeMax } {
   set STAT_VolumeMax $vol;
  }
  if { $vol < $STAT_VolumeMin } {
   set STAT_VolumeMin $vol;
  }
  if { $key > $STAT_KeyMax } {
   set STAT_KeyMax $key;
  }
  if { $key < $STAT_KeyMin } {
   set STAT_KeyMin $key;
  }
 }
 return;
}

proc print_statistics {} {
# schreibt Spielstatistik auf Standardausgabe
 global STAT_NumberOfHits\
        STAT_DurationTotal STAT_DurationMax STAT_DurationMin\
        STAT_VolumeTotal   STAT_VolumeMax   STAT_VolumeMin\
        STAT_KeyTotal      STAT_KeyMax      STAT_KeyMin;
 set num [expr double($STAT_NumberOfHits)];
 puts "Anschlagstatistik:"
 puts [format "- Gesamtzahl Anschlaege    = %11d times" $STAT_NumberOfHits];
 puts [format "- minimale Anschlagdauer   = %11.5f sec" $STAT_DurationMin];
 puts [format "- maximale Anschlagdauer   = %11.5f sec" $STAT_DurationMax];
 if { $STAT_NumberOfHits != 0 } {
  set average [expr $STAT_DurationTotal/$num];
 } else {
  set average 0;
 }
 puts [format "- mittlere Anschlagdauer   = %11.5f sec" $average];
 puts [format "- minimale Anschlagstaerke = %11.5f dB" $STAT_VolumeMin];
 puts [format "- maximale Anschlagstaerke = %11.5f dB" $STAT_VolumeMax];
 if { $STAT_NumberOfHits != 0 } {
  set average [expr $STAT_VolumeTotal/$num];
 } else {
  set average 0;
 }
 puts [format "- mittlere Anschlagstaerke = %11.5f dB" $average];
 set cps  [expr 440.0*pow(2,([expr $STAT_KeyMin - 21]/12.0))];
 puts [format "- minimale Tastennummer    = %11.5f = %9.3f Hz" $STAT_KeyMin $cps];
 set cps  [expr 440.0*pow(2,([expr $STAT_KeyMax - 21]/12.0))];
 puts [format "- maximale Tastennummer    = %11.5f = %9.3f Hz" $STAT_KeyMax $cps];
 if { $STAT_NumberOfHits != 0 } {
  set average [expr $STAT_KeyTotal/$num];
 } else {
  set average 0;
 }
 set cps  [expr 440.0*pow(2,([expr $average - 21]/12.0))];
 puts [format "- mittlere Tastennummer    = %11.5f = %9.3f Hz" $average $cps];
 return;
}

proc print_sys_info out {
# Gibt verschiedene System-Informationen aus
 global tcl_patchLevel tcl_platform VS;
 puts $out "; Created with Tcl-Music-Script \"music.tcl\" version $VS";
 puts $out "\n; *** Begin of System Information ***"
 puts $out "; tcl_patchLevel: $tcl_patchLevel";
 puts $out "; tcl_platform:"
 foreach {key value} [array get tcl_platform] {
  puts $out [format "; %12s %s" $key $value];
 }
 puts $out "; *** End of System Information ***\n";
}

proc get_name_id { rCTX name domain } {
# Ordnet einem Namen eine eindeutige Nummer zu bzw. gibt eine bereits
# zugeordnete Nummer zurueck. Ueber 'domain' koennen verschiedenen Bereiche
# angesprochen werden, z.B. "instr" fuer Instrumente, "ftabs" fuer
# Funktionstabellen.
 global $rCTX;
 set name_list [TREEget $rCTX $domain];
 set i 0;
 foreach n $name_list {
  incr i;
  if { [string compare $name $n] == 0 } {
   return $i;
  }
 }
 incr i;
 lappend name_list $name;
 TREEset $rCTX $domain $name_list;
 return $i;
}

# ----------------------------------------------------------------------------
puts "Lektion x : Umrechnungen der Liedeintraege, song-entry conversions";
# ----------------------------------------------------------------------------

proc get_btbr rCTX {
# liefert Anzahl der Schlaege pro Takt (btbr = beats per bar)
 global $rCTX;
 set tmsg [TREEget $rCTX tmsg];
 if { [scan $tmsg "%d/%d" z n] != 2 } {
  puts "*** falsche Taktangabe $tmsg ***";
  set z 4;
 }
 return $z;
}

proc get_bttv rCTX {
# liefert 1/Notenwert pro Schlag (bttv = 1/time value per beat)
 global $rCTX;
 set tmsg [TREEget $rCTX tmsg];
 if { [scan $tmsg "%d/%d" z n] != 2 } {
  puts "*** falsche Taktangabe $tmsg ***";
  set n 4;
 }
 return $n;
}

proc get_tmbt rCTX {
# liefert Zeit pro Schlag (tmbt = time per beat)
 global $rCTX;
 set tmtv [TREEget $rCTX tmtv];
 set bttv [get_bttv $rCTX];
 set tmbt [expr $tmtv/$bttv]; 
 return $tmbt;
}

# ----------------------------------------------------------------------------
puts "Lektion x : Takt und Rhythmus, time and rhythm";
# ----------------------------------------------------------------------------

proc clock_get_time rCTX {
# aktuelle Zeit holen
 global $rCTX;
 set crtm [TREEget $rCTX crtm];
 # maximalen Zeitwert aktualisieren (wichtig z.B. bei Rueckwaertsspiel)
 if { $crtm > [TREEget $rCTX tttm] } {
  TREEset $rCTX tttm $crtm;
 }
 return $crtm;
}

proc clock_set_time { rCTX crtm } {
# Aktuelle Zeit in Sekunden setzen und die Differenz zur
# ursrpruenglichen Zeit zurueckgeben.
 global $rCTX;
 set dura [expr $crtm - [clock_get_time $rCTX]];
 TREEset $rCTX crtm $crtm;
 # maximalen Zeitwert aktualisieren
 if { $crtm > [TREEget $rCTX tttm] } {
  TREEset $rCTX tttm $crtm;
 }
 return $dura;
} 

proc clock_get_stamp_tm { rCTX } {
# holt den Zeitstempel in Sekunden
 global $rCTX;
 return [TREEget $rCTX sttm];
}

proc clock_get_stamp_tv { rCTX } {
# holt den Zeitstempel in Notenwerten
 global $rCTX;
 return [TREEget $rCTX sttv];
}

proc clock_set_stamp { rCTX } {
# setzt den Zeitstempel auf die aktuelle Zeit und den aktuellen 
# Notenwert und gibt die aktuelle Zeit in Sekunden aus.
 global $rCTX;
 TREEset $rCTX sttv [TREEget $rCTX crtv];
 set crtm [clock_get_time $rCTX];
 TREEset $rCTX sttm $crtm;
 return $crtm;
}

proc clock_get_tval rCTX {
# aktuelle Zeit als Notenwert holen
 global $rCTX;
 return [TREEget $rCTX crtv];
}

proc clock_set_tval { rCTX tval } {
# Aktuelle Zeit als Notenwert setzen und die Differenz zur
# ursrpruenglichen Zeit zurueckgeben.
 global $rCTX;
 set nval [expr $tval - [clock_get_tval $rCTX]];
 TREEset $rCTX crtv $tval;
 return $nval;
} 

proc time_function { rCTX nval rtval rtime rdura } {
# Standard-Zeitfunktion
 global $rCTX;
 upvar $rtval tval $rtime time $rdura dura;
 # aktuelle Zeit in Sekunden und als Notenwerk merken
 set crtm [clock_get_time $rCTX];
 set crtv [clock_get_tval $rCTX];
 # beide Werte neu berechnen
 set tval [expr $crtv + $nval];
 set time [expr $crtm + ($nval*[TREEget $rCTX tmtv]\
           *pow([TREEget $rCTX vstp],-[TREEget $rCTX velo]))];
 # Spieldauer in Sekunden berechnen
 set dura [expr $time - $crtm];
 return ;
}

proc clock_advance { rCTX nval } {
# Zeit um Notenwert fortschreiben und die Zeitdifferenz zurueckgeben
 global $rCTX;
 # aktuelle Zeit in Sekunden und als Notenwerk merken
 set crtm [clock_get_time $rCTX];
 set crtv [clock_get_tval $rCTX];
 # neue Werte und Spieldauer berechnen
 [TREEget $rCTX tmfn] $rCTX $nval tval time dura;
 # neu berechnete Werte in Klangbaum eintragen
 clock_set_time $rCTX $time;
 clock_set_tval $rCTX $tval;
 # Spieldauer weitergeben
 return $dura;
}

proc beat_advance { rCTX crtv nval crtm dura } {
 global $rCTX;
 set beats [list];
 set bttv  [get_bttv $rCTX];    # Schlaege pro Notenwert
 set btbr  [get_btbr $rCTX];    # Schlaege pro Takt
 # aktueller Schlagindex (double)
 set crbt  [expr [TREEget $rCTX btbs]+double($crtv)*$bttv];
 set dbt   [expr $nval*$bttv];  # Dauer in Schlaegen (double)
 set tmbt  [expr $dura/$dbt];   # Zeit pro Schlag
 set bte   [expr $crbt + $dbt]; # Endschlagindex (double)
 set bt0   [expr ceil($crbt)];  # naechster ganzzahliger Schlagindex
 for { set bt $bt0; set tm [expr $crtm+($bt0-$crbt)*$tmbt] }\
     { $bt < $bte } { set bt [expr $bt+1.0]; set tm [expr $tm+$tmbt] } {
  set br [expr int($bt/$btbr)];
  set ht [expr $bt-$br*$btbr];
  lappend beats $tm; # Schlag-Zeitpunkt
  lappend beats [expr int($bt)]; # Schlagindex
  lappend beats [expr int($br)]; # Taktindex
  lappend beats [expr int($ht)]; # Schlag innerhalb Takt
 }
 return $beats;
}

# ----------------------------------------------------------------------------
puts "Lektion x : Artikulation, expression ";
# ----------------------------------------------------------------------------

proc off_volume { rCTX pifo crespos } {
# Modifikation der Anschlagstaerke
 global $rCTX;
 return [[TREEget $rCTX vlfn] $rCTX $pifo $crespos];
}

proc volume_function { rCTX pifo crespos } {
# Standard Anschlagstaerkefunktion 
# pifo    = Piano/Forte-Zaehler
# crespos = relative Position innerhalb eines Crescendo-Bereichs (0 bis 1)
 global $rCTX;
 return [expr [TREEget $rCTX vlbs]+$pifo*[TREEget $rCTX fstp]+\
              [TREEget $rCTX cres]*$crespos];
}

proc off_key { rCTX key pos } {
# Modifikation der Tastennummer
 global $rCTX;
 return [[TREEget $rCTX kyfn] $rCTX $key $pos];
}

proc key_function { rCTX key pos } {
# Standard Tastennummerfunktion
# key = Originaltastennummer
# pos = relative Position innerhalb eines Taktes (0 bis 1)
 global $rCTX;
 return $key;
}

# ----------------------------------------------------------------------------
puts "Lektion x : Notenlesen, read music";
# ----------------------------------------------------------------------------

# 1. Vorauswahl der Notation

source "basics/notes_123.tcl"; # wir lernen 123-Notation

# 2. Prozeduren unabhaengig von der Notation

proc get_key_by_letter letter {
 switch -glob -- [string index $letter 0] {
  "C" {set k  0} "D" {set k  2} "E" {set k  4} "F" {set k  5}
  "G" {set k  7} "A" {set k  9} "B" {set k 11} "H" {set k 11}
  "c" {set k 12} "d" {set k 14} "e" {set k 16} "f" {set k 17}
  "g" {set k 19} "a" {set k 21} "b" {set k 23} "h" {set k 23}
  "*" {set k P}  
 }
 if { $k != "P" } {
  switch -glob -- [string index $letter 1] {
   "#" { incr k    }
   "b" { incr k -1 } 
  }
 }
 return $k;
}

proc get_key_by_ratio ratio {
 global X12LN2;
 if { [scan $ratio "%d:%d" z n] == 2 && $z != 0 && $n != 0} {
  set k [expr $X12LN2*log(double($z)/double($n))];
 } else {
  set k "P";
 }
 return $k;
}

proc get_key_by_freq freq {
 global CPS0 X12LN2;
 return [expr $X12LN2*log(double($freq)/$CPS0)]
}

# Metadaten = Zeichenfolge mit Dynamik- und Artikulations-Angaben
# f = forte   = eine Stufe lauter
# p = piano   = eine Stufe leiser
# v = veloce  = eine Stufe schneller
# a = adagio  = eine Stufe langsamer
# _ = legato  = Beginn oder Ende eines Legatobogens
# * = fermate = haelt einzelne Note um eine Stufe laenger

proc get_meta_string { indx metas } {
 # holt Meta-Daten fuer eine Note aus der Metalinie bzw. ".", wenn der
 # Notenindex ausserhalb der Liste ist.
 set meta [lindex $metas $indx];
 if { $meta == "" } {
  set meta ".";
 }
 return $meta; 
}

proc parse_meta_info { meta lega fort velo ferm} {
# analysiert Meta-Daten und fuellt drei Variablen mit dem Ergebnis:
# lega = 1 fuer Legato-Bogen gesetzt, 0 sonst
# fort = piano/forte Lautstaerke-Aenderung
# velo = adagio/veloce Geschwindigkeits-Aenderung 
 upvar $lega lg $fort ft $velo vl $ferm fm;
 set lg [string match "*_*" $meta];
 set ft [expr [regexp -all "f" $meta] - [regexp -all "p" $meta]];
 set vl [expr [regexp -all "v" $meta] - [regexp -all "a" $meta]];
 set fm [expr [regexp -all {[\*]} $meta]];
 return $meta;
}

proc pass_note_melody { rCTX indx crtv nval nkey meta csym } {
# Diese Prozedur spielt die Melodie ("pass_note_rhythm" spielt Rhythmus).
 return [pass_note $rCTX $indx $crtv $nval $nkey\
         $meta $csym "instr_play_note"];
}

proc pass_note_rhythm { rCTX indx crtv nval nkey meta csym } {
# Diese Prozedur spielt den Rhythmus ("pass_note_melody" spielt Melodie).
 return [pass_note $rCTX $indx $crtv $nval $nkey\
         $meta $csym "instr_beat_note"];
}

proc pass_note { rCTX indx crtv nval nkey meta csym play_proc } {
# Noten-Uebergabe-Prozedur, die von der Notenleseprozedur "read_notes"
# (s. notes_123.tcl, notes_abc.tcl) zur Notenuebergabe aufgerufen wird.
 global $rCTX;
 set lg 0;
 # --- nach Index verzweigen ---
 switch -glob -- $indx {
  "begin" {
   instr_begin_notes $rCTX;
  }
  "end" {
   instr_end_notes $rCTX;
   # Buchhaltung zuruecksetzen
   set ${rCTX}(velo) 0;  # adagio/veloce-Stufe
   set ${rCTX}(fort) 0;  # piano/forte-Stufe
   set ${rCTX}(lega) 0;  # Legato-Kennung
   set ${rCTX}(csym) {}; # Akkord-Symbol
  }
  "*" {
   # --- Akkord-Symbol uebertragen ---
   if { $csym != "." } { 
    set ${rCTX}(csym) $csym;
   }
   # --- Meta-Daten auswerten ---
   parse_meta_info $meta lg ft vl fm;
   if { $lg && ![TREEget $rCTX lega] } {
    set ${rCTX}(lega) 1;
    set lg 0;
   } 
   set ${rCTX}(fort) [expr [TREEget $rCTX fort]+$ft];
   set ${rCTX}(velo) [expr [TREEget $rCTX velo]+$vl];
   
   # --- aktuellen Zeitpunktholen und Zeit fortschreiben ---   
   set when [clock_get_time $rCTX];
   set frmt [TREEget $rCTX frmt];
   if { $fm > 0 && $frmt > 0 } {
    set dura [clock_advance $rCTX [expr $nval*(1.0+$fm*$frmt)]];
   } else {
    set dura [clock_advance $rCTX $nval];      
   }
   # --- Zeitkorrektur fuer Rueckwaertsspiel (negative Dauer) ---
   if { $dura < 0 } {
    set when [expr $when+$dura];
    set dura [expr -$dura];
   }
      
   # --- Tastennummer korrigieren ---
   set key  [off_key $rCTX $nkey 0];

   # --- weitere Spielparameter holen ---
   # inum = Instrumentennummer,
   # volm = Lautstaerke,
   # loud = Lautstaerkefaktor
   set inum [register_inam $rCTX [TREEget $rCTX inam]];
   set volm [off_volume $rCTX [TREEget $rCTX fort] 0];
   set loud [TREEget $rCTX loud];

   # --- Note spielen ---   
   $play_proc $rCTX 0 $inum $crtv $nval $when $dura $volm $loud $key $meta $csym;
   
   # --- Statistik mitschreiben ---
   register_statistics $rCTX $dura $volm $key;
  }
 }
 # --- Gesamtnotenwert fortschreiben ---
 set crtv [expr $crtv + $nval];
 # --- eventuell Legato-Kennung ausschalten ---
 if { $lg && [TREEget $rCTX lega] } {
  set ${rCTX}(lega) 0;
 } 
 return $crtv;
}

proc join_words words {
# Notentext fuer Anzeige zusammensetzen
 set result [join [eval "list $words"]];
 regsub -all {\- } $result {} result;
 return $result;
}

proc csound_comment { line } {
# verteilt zu lange Csound-Kommentarzeilen auf mehrere Zeilen
 regsub -all ".{80}(?=.+?)" $line "&\n; " line;
 return "; $line"; 
}

proc resolve_music_pattern { rCTX pattern } {
# Aufloesung von Wiederholungs- und Ersetzungsmustern
 global $rCTX;
#puts "*** Original: $pattern ***";
 set result "";
 # Muster rekursiv aufloesen bis sich nichts mehr aendert
 while { 1 } {
  # Kontext-Eintraege ersetzen, i.e. <name> durch <wert>
  regsub -all {<([a-zA-Z]+[0-9a-zA-Z]*)>} "$pattern" {<[TREEget $rCTX \1]>} result;
  set result [subst $result];
#puts "*** <name> ersetzt: $result ***";
  # Muster ohne Wiederholungszaehler aufloesen 
  regsub -all {<([^<>]*)>([^0-9]|$)} "$result" {\1\2} result
#puts "*** <xxxx> ersetzt: $result ***";
  # Muster mit Wiederholungszaehler aufloesen
  regsub -all {<([^<>]*)>([0-9]+)} "$result" {[string repeat {\1} \2]} result;
  set result [subst $result];
#puts "*** <xxxx>N ersetzt: $result ***";
  if { $result == $pattern } break;
  set pattern $result;
 }
 # zum Schluss noch alle Leerzeichen entfernen
 regsub -all {[[:space:]]+} "$result" {} result; 
#puts "*** Leerzeichen entfernt: $result ***";
 # fertig
 return $result;
}

proc parse_chord {rCTX csym cmod kcllst } {
# analysiert Akkordsymbol und fuellt entsprechende Variablen
# csym  = Akkordsymbole f. absteigende Oktaven (durch '/' getrennt)
# cmod  = allgemeine Modusangaben f. absteigende Oktaven
# kclst = Liste von Tastennummern-/Verzoegerungslisten fuer Akkorde
 upvar $kcllst kcll;
 global $rCTX;
 # Vorbelegungen
 set cstc [TREEget $rCTX cstc];
 set kcll [list]; set kbase 0; set koct P;
 # Bildung der Listen f. absteigende Oktaven
 regsub -all {/} $csym "\} 12 \{" csym;
 regsub -all {[|]} $csym "\} 0 \{" csym;
 set csym "0 \{$csym\}";
 eval "set symlst [list $csym]";  
 set modlst [split $cmod "/|"]; 
 # Schleife ueber Oktaven
 set kybs 0;
 foreach { key sym } $symlst mod $modlst {
  if { $key == "" } continue;
  # Grundton aktualisieren
  set kybs [expr $kybs-$key];
  # Akkordsymbol auswerten:
  # 1. symbolischer Grundakkord
  if { [regexp {(^bb)|(^[a-hA-H][bbm]?)} ${sym} next] } {
   regsub {(^bb)|(^[a-hA-H][bbm]?)} ${sym} {} sym;
   set koct [get_key_by_letter $next]; 
   set next [string index ${sym} 0];
   if { [string is integer $koct] } {
    set kbase $koct;
    # 1.1. vermindert, Moll, Dur, uebermaessig
    set k0 $kbase; set k1 $k0; set k2 $k0;
    if       { $next == "m" } {
     incr k1 3; incr k2 7; 
    } elseif { $next == "+" } {
     incr k1 4; incr k2 8; 
    } elseif { $next == "0" } {
     incr k1 3; incr k2 6; 
    } else {
     incr k1 4; incr k2 7; 
    }
    if { $next == "m" || $next == "+" || $next == "0" } {
     set sym [string range ${sym} 1 end]; 
     set next  [string index ${sym} 0];
    }
    # 1.2 Dominantseptakkord
    if { $next == 7 } {
     incr k0 7; incr k1 7; incr k2 7; set k3 $k2; incr k3 3;
     set kcl [list $k0 $k1 $k2 $k3];
     set sym [string range ${sym} 1 end]; 
     set next  [string index ${sym} 0];
    } else {
     set kcl [list $k0 $k1 $k2];
    }
   }
  }
  # 2. weitere Toene hinzufuegen
  if { [regexp {(\()(.*)(\))} ${sym} more] } {
   regsub {(\()(.*)(\))} ${sym} {} sym;
   set more [string range $more 1 [expr [string length $more]-2]];
   foreach k [split $more {,}] {
    if { [string is double $k] } {
     lappend kcl $k;
    } elseif { [regexp {^[1-9][0-9]*:[1-9][0-9]*$} $k] } {
     set knew [get_key_by_ratio $k];
     lappend kcl $knew;
    } else {
     set knew [get_key_by_letter $k];
     if { [string is integer $knew] } {
      lappend kcl $knew;
     } else {
      puts "*** ungueltige Tastenkennung \"$k\"";
     }
    }
   }
   if { ![string is integer $koct] && [llength $kcl] > 0 } {
    set kbase [lindex $kcl 0];
   }
  }
  # 3. wenn keine Toene, letzten Grundton uebernehmen
  if { [llength $kcl] == 0 } {
   set kcl [list $kbase]; 
  }
  # 4. Umkehrungen
  if { [scan ${sym} "%d" num] == 1 } {
   set len [llength $kcl];
   if { $num > 0 } {
    for { set i 0 } { $i < $num } { incr i } {
     set index [expr $i%$len];
     set knew [expr [lindex $kcl $index]+12];
     set kcl [lreplace $kcl $index $index $knew];
    }
   } elseif { $num < 0 } {
    for { set i 0 } { $i > $num } { incr i -1 } {
     set index [expr $len-(-$i%$len)-1];
     set knew [expr [lindex $kcl $index]-12];
     set kcl [lreplace $kcl $index $index $knew];
    }
   }
  }  
  set len [llength $kcl];

  # Akkordmodus auswerten:
  # 1. Kennungen fuer sortieren, revertieren und Pedal herausziehen,
  #    anschliessend Gueltigkeit des Akkordmodus sicherstellen
  set sort [string match "*s*" ${mod}];
  set rvrt [string match "*r*" ${mod}];
  set peda [string match "*u*" ${mod}];
  # nicht interpretierte Zeichen entfernen
  regsub -all {[^fp\!\=\_\,\.\;\:\'\*]+} ${mod} {} mod;
  # leeren Akkordmodus vermeiden
  if { ${mod} == "" } {
   set mod [string repeat "," $len];
  }
  # Zeichen ohne Schlagzeit am Ende vermeiden (wenigsten ";" anhaengen)
  if { [regexp {[fp\,\'\*]+} [string index ${mod} end]] } {
   set mod "${mod};";
  }

  # 2. sortieren
  if { $sort } {
   set kcl [lsort -integer -increasing $kcl];
  }

  # 3. Oktavenkorrektur und eventuell revertieren
  set kclnew [list];
  if { $rvrt } {
   set insi 0;
  } else {
   set insi end;
  }
  for { set i 0 } { $i < $len } { incr i } {
   set knew [expr [lindex $kcl $i]+$kybs];
   set kclnew [linsert $kclnew $insi $knew]; 
  }

  # 4. arpeggieren und dynamisieren, d.h. Verzoegerungsliste erstellen,
  # Pausen einfuegen, piano/forte-Zaehlerliste erstellen
  set dlclst [list];
  set stclst [list];
  set piflst [list];
  set pifo 0;
  set lmod [string length ${mod}];
  set nmbt [regexp -all {[\!\=\_\.\;\:]} ${mod}];
  if { $nmbt <= 0 } { set nmbt 1 };
  set dlcx [expr 1.0/$nmbt];
  set i 0; set j 0; set dlc 0; set kcl [list];
  while { $j < $lmod } {
   set next [string index ${mod} $j];
   incr j;
   if { $i < $len } {
    if { [regexp {[\!\=\_\.]} $next] } {
     lappend kcl [lindex $kclnew $i] 
     lappend piflst $pifo;
     lappend dlclst $dlc;
     set dlc [expr $dlc+$dlcx];
     switch -exact -- $next {
      "." {lappend stclst 0}
      "!" {lappend stclst 1}
      "=" {lappend stclst 2}
      "_" {lappend stclst 3}
     } 
     incr i;
    } elseif { $next == "," } {
     lappend kcl [lindex $kclnew $i] 
     lappend piflst $pifo;
     lappend dlclst $dlc;
     lappend stclst 0;
     incr i;
    } elseif { $next == "'" } {
     incr i; 
    } elseif { $next == "*" } {
     incr i [expr $len - 1];
    } elseif { $next == "f" } {
     incr pifo;
    } elseif { $next == "p" } {
     incr pifo -1;
    }
    set i [expr $i%$len];
   }
   if { $next == ";" } {
    set dlc [expr $dlc+$dlcx];
   } elseif { $next == ":" } {
    lappend kcl "P"; 
    lappend piflst $pifo;
    lappend dlclst $dlc;
    lappend stclst 0;
    set dlc [expr $dlc+$dlcx];
   }
  }
  lappend dlclst 1.0;
  set len [llength $kcl];

  # 5. Dauernliste erstellen
  set cstcl [list 0 [expr -$cstc*$dlcx]\
   [expr -$cstc*$dlcx/2.0] [expr $cstc*$dlcx/2.0]]; 
  set i 0; set cnt 1; set durlst [list];
  while { $i < $len } {
   set dlc0 [lindex $dlclst $i];
   set stc  [lindex $cstcl [lindex $stclst $i]];
   incr i;
   set dlc1 [lindex $dlclst $i];
   if { $peda } {
    set dur [expr 1.0-$dlc0];
   } else {
    set dur [expr $dlc1-$dlc0];
   }
   if { $dur > 0.0 || $i >= $len } {
    for { set j 0 } { $j < $cnt } { incr j } {
     lappend durlst [expr $dur+$stc];
    }
    set cnt 1;
   } else {
    incr cnt;
   }  
  }

  # 6. Listen mischen
  set kclnew [list];
  for { set i 0 } { $i < $len } { incr i } {
   lappend kclnew [lindex $kcl $i];
   lappend kclnew [lindex $dlclst $i];
   lappend kclnew [lindex $durlst $i];
   lappend kclnew [lindex $piflst $i];
  }
  # Ergebnis hinzufuegen
  lappend kcll $kclnew;
  set kcl [list];
 }
}

proc N-line line {
# Notenlinien-Prozedur wird aus Partitur aufgerufen !
 upvar 1 rCTX rCTX;
 global $rCTX;
 set out [TREEget $rCTX out]; 
 # --- komplettes Zeilenfeld in Kontext kopieren ---
 array set $rCTX [concat $line];
 # --- Textzeile ausgeben ---
 puts $out "; [join_words [TREEget $rCTX L]]";
 # --- das zu spielende Instrument lernen ---
 learn_instrument $rCTX [TREEget $rCTX inam];
 # --- Noten spielen ---
 puts [join_words [TREEget $rCTX L]];
 read_notes $rCTX "pass_note_[TREEget $rCTX pmod]";
}

# ----------------------------------------------------------------------------
puts "Lektion x : Csound-Instrumente spielen, playing Csound instruments";
# ----------------------------------------------------------------------------

# Das Spielen eines bestimmten Instrumentes lernen wir spaeter, hier gibt es
# nur die allgemeinen Vorkenntnisse fuer das Spielen von Csound-Instrumenten 

proc store_i_statement { rCTX args } {
# speichert Parameter fuer ein Csound i-Statement zwischen,
# ueberschreibt alle vorher gespeicherten Parameter,
# loeschen durch Aufruf ohne weitere Parameter.
 global $rCTX;
 array unset $rCTX "iPar*";
 set i 1;
 foreach iPar $args {
  set ${rCTX}(iPar[format "%02d" $i]) $iPar;
  incr i;
 }
}

proc flush_i_statement rCTX {
# gibt gespeichertes i-Statement an den Ausgabekanal weiter
# und loescht dasselbe anschliessend
 global $rCTX;
 set pars [lsort [array names $rCTX "iPar*"]];
 set prefix "";
 set suffix "";
 foreach par $pars {
  puts -nonewline [TREEget $rCTX out] "$prefix[set ${rCTX}($par)]";
  set prefix " ";
  set suffix "\n";
 }
 puts -nonewline [TREEget $rCTX out] "$suffix";
 array unset $rCTX "iPar*";
}

proc learn_instrument { rCTX inam } {
# falls notwendig, ein bestimmtes Instrument spielen lernen
# gibt das urspruenglich gelernte Instrument zurueck
 global $rCTX;
 set crin [TREEget $rCTX crin]; 
 if { $crin != $inam } {
  source "instruments/$inam.tcl";
  TREEset $rCTX crin $inam;
 }
 return $crin;
}

# ----------------------------------------------------------------------------
puts "Lektion x : Partituren und -Orchester, scores and orchestras";
# ----------------------------------------------------------------------------

proc register_inam { rCTX inam } {
# Weist einem Instrumentennamen eine eindeutige Nummer (ID) zu.
# Rueckgabewert: die zugewiesene Nummer.
 global $rCTX;
 return [get_name_id $rCTX $inam "instr"];
}

proc register_fstat { rCTX inum fstat } {
# Weist einem f-Statement eine eindeutige Nummer (ID) zu. Die eigentlichen
# Tabellendaten und die Tabellennummer werden dabei getrennt gespeichert.
 global $rCTX;
 set fnum 0;
 # --- f-Statement pruefen und die urspruengliche Tabellennummer merken ---
 if { [scan $fstat "f %d" iftnum] == 1 && $iftnum != 0 } {
  # --- f-Statement von Tabellennummer befreien --- 
  regsub {f[\ ]+[0-9]+} $fstat "f %d" fst0; 
  # --- ID fuer den Rest zuweisen (neue Tabellennummer) ---
  set fnum [get_name_id $rCTX $fst0 "fstats"];   
  # --- die zugewiesene ID unter einem Namen speichern, der eindeutig aus
  #     Instrumentennummer und urspruenglicher Tabellennumer gebildet wird ---
  set fnums [TREEget $rCTX fnums];
  lappend fnums ft$iftnum-$inum $fnum;
  TREEset $rCTX fnums $fnums;
 }
 return $fnum;
}

proc place_instrument { rCTX out iname inum } {
# registriert Funktionstabellen und schreibt die Csound ORC-Datei
 # --- erstmal alle f-Statements aus der SCO-Datei registrieren ---
 global $rCTX;
 set in [open "instruments/$iname\.sco" "r"];
 while { [gets $in line] != -1 } {
  register_fstat $rCTX $inum $line;
 }
 close $in;
 # --- ORC-Datei schreiben ---
 set in [open "instruments/$iname\.orc" "r"];
 array set fnums [TREEget $rCTX fnums]; # neue Funktionstabellennummern
 while { [gets $in line] != -1 } {
  # --- skip header entries ---
  if { [regexp {^sr|^kr|^ksmps|^nchnls} $line] } continue;
  # --- Instrumentennummer in i-Statement ersetzen ---
  if { [regexp {instr [0-9]+} $line] } {
   puts $out "instr $inum ; *** instrument number replaced by music script ***";
   continue;
  }
  # --- Funktionstabellennummer ersetzen ---
  if { [regexp {ift[0-9]+[\ ]+=[\ ]+[0-9]+.*} $line] } {
   if { [scan $line "%s = %d" iftnam iftnum] != 2 } {
    error "*** ungueltige FT-Zuweisung \"$line\" ($iname) ***";
   }
   set fnam "ft$iftnum-$inum";
   set fnum [set fnums($fnam)];
   if { $fnum != "" } {
    puts $out "$iftnam = $fnum ; *** function table number replaced by music script ***";
   } else {
    error "*** unerlaubte FT-Zuweisung \"line\" ($iname) ***";
   }
   array unset fnums $fnam; # einmal zugewiesenen Nummer vergessen
   continue;
  }
  # --- any other lines to be transfered without changes ---
  puts $out $line;
 }
 close $in;
 # --- vom Instrument nicht benutzte Funktionstabellen loeschen ---
 array set fused [TREEget $rCTX fnums];
 foreach name [array names fnums] {
  if { [string match "ft*-$inum" $name] } {
   array unset fused $name;
  }
 }
 TREEset $rCTX fnums [array get fused];
}

proc write_orc_file rCTX {
 global $rCTX, tcl_platform;
 set fnam [TREEget $rCTX loca];
 # open ORC file
 puts "+ open ORC file \"csound/$fnam.orc\""
 set out [open "csound/$fnam.orc" "w"];
 # write header
 print_sys_info $out;
 puts $out "; *** header entries replaced by music script ***";
 puts $out "sr     = 44100";
 puts $out "kr     = 11025";
 puts $out "ksmps  =     4";
 puts $out "nchnls =     2";
 puts $out "; globals\n"; # a place where to insert additional global variables
 # place instruments
 set i 0;
 foreach n [TREEget $rCTX instr] {
  incr i;
  place_instrument $rCTX $out $n $i;
 }
 # close ORC file
 puts "- close ORC file \"csound/$fnam.orc\"";
 close $out;
 if { [string compare $tcl_platform(platform) "macintosh"] == 0 } {
  file attributes "csound/${fnam}.orc" -type TEXT -creator VRmi;
 }
}

proc open_sco_file rCTX { 
# oeffnet Csound-SCO-Datei und schreibt Kopfeintraege
 global $rCTX CP;
 set fnam "csound/[TREEget $rCTX loca].sco";
 set out [open $fnam "w"];
 puts "+ open SCO file \"$fnam\"";
 TREEset $rCTX out $out;
 puts $out $CP;
 print_sys_info $out;
 puts $out "; --- i-Statements ---\n"; # where to place signal output mixer i-istatement
 return $out;
} 

proc close_sco_file rCTX {
 global $rCTX, tcl_platform;
 set out [TREEget $rCTX out];
 # Funktions-Tabellen schreiben
 puts $out "\n; --- f-Statements ---";
 set i 0;
 set fnums [TREEget $rCTX fnums];
 foreach fstat [TREEget $rCTX fstats] {
  incr i;
  if { [lsearch -exact $fnums $i] >= 0 } { 
   puts $out [format $fstat $i];
  }
 }
 # Ende schreiben
 puts $out "\n; --- Ende ---";
 puts $out [format "f0 %.2f ; total time control" [TREEget $rCTX tttm]];
 puts $out "e ; Ende";
 close $out;
 set ${rCTX}(out) {};
 puts "- close SCO file \"csound/[TREEget $rCTX loca].sco\"";
 set fnam "csound/[TREEget $rCTX loca].sco";
 if { [string compare $tcl_platform(platform) "macintosh"] == 0 } {
  file attributes $fnam -type TEXT -creator ttxt;
 }
 print_statistics;
 return {}
}

# ----------------------------------------------------------------------------
puts "Lektion x : Spielprozeduren, play-procedures";
# ----------------------------------------------------------------------------
proc play_closing {} {
# Abschlussprozedur
 upvar 1 rCTX rCTX;
 clock_advance $rCTX [TREEget $rCTX cltv];
 return;
}

proc play_opening {} {
# Eroeffnungsprozedur
 upvar 1 rCTX rCTX;
 clock_advance $rCTX [TREEget $rCTX optv];
 return;
}

proc play_pre {} {
# Vorspielprozedur
 upvar 1 rCTX rCTX;
 clock_advance $rCTX [TREEget $rCTX prtv];
 return;
}

proc play_post {} {
# Nachspielprozedur
 upvar 1 rCTX rCTX;
 clock_advance $rCTX [TREEget $rCTX pstv];
 return;
}

proc play_pause nval {
# spielt eine Pause
 upvar 1 rCTX rCTX;
 clock_advance $rCTX $nval;
}

proc play_chord {} {
# spielt Akkorde
 upvar 1 rCTX rCTX;
 # Anschlag- und Zeitvariablen holen
# puts "***** play_chord time stamp = [TREEget $rCTX sttm], [TREEget $rCTX sttv]";
 set bthk  [TREEget $rCTX bthk];
 set inam  [TREEget $rCTX inam];
 set loud  [TREEget $rCTX loud];
 set chtv  [TREEget $rCTX chtv];
 set fstp  [TREEget $rCTX fstp];
 set text  [TREEget $rCTX text];
 set out   [TREEget $rCTX out];
 # Akkordsymbole holen und aufbereiten
 set csym [resolve_music_pattern $rCTX "[TREEget $rCTX csym]"];
 set cmod [resolve_music_pattern $rCTX "[TREEget $rCTX cmod]"];
 set chbt [resolve_music_pattern $rCTX "[TREEget $rCTX chbt]"];
 puts $out "; song = \"[TREEget $rCTX song]\"";
 puts $out [csound_comment "csym = \"$csym\""];
 puts $out [csound_comment "cmod = \"$cmod\""];
 puts $out [csound_comment "chbt = \"$chbt\""];
 if { $text != "" } { puts $out [csound_comment "text = \"$text\""] }
 puts -nonewline $out "; instr  time     dura   volume";
 puts $out "            pitch     loud dist azim size angl   note-ID  TV-time  TV-dura";
 # und auswerten
 set chbtl [split $chbt "/|"];
 set lbtl [llength $chbtl];
 set inum  [register_inam $rCTX $inam];
 learn_instrument $rCTX $inam;
 instr_begin_notes $rCTX;
 parse_chord $rCTX $csym $cmod kclst;
 # Akkorde spielen
 set iocn 0;                      # Oktavierunszaehler
 set crtm [clock_get_time $rCTX]; # Realzeit merken
 set crtv [clock_get_tval $rCTX]; # Notenzeit merken
 foreach kcl $kclst {
  # replace empty chord beat by a single beat
  set chbt  [lindex $chbtl $iocn];
  if { $chbt == "" } {
   set chbt "."
  }
  # Piano/Forte-Zaehler zuruecksetzen
  set pifo 0;
  # neuer Akkord-Notenwert durch Akkordverlaengerung
  set chtvb [expr $chtv*([regexp -all {_} $chbt]+1)];
  # Schlagzahl und neuer Akkord-Notenwert durch Anschlagteilung
  set nmbt [regexp -all {[\.\;\:]} $chbt];
  if { $nmbt > 0 } {
   set chtv1 [expr double($chtvb)/$nmbt];
  } else {
   set chtv1 $chtvb;
  }
  # erst Spielzeitliste (in Notenwerten) in einem Vorlauf erstellen
  for { set j 0; set tvlst0 [list]; set next [string index $chbt 0]; set k 0; }\
      { $next != "" } { incr j; set next [string index $chbt $j] } {
   if { $next == "." || $next == ":" } {
    lappend tvlst0 [expr $k*$chtv1];
    incr k;
   } elseif { $next == ";" } {
    incr k;
   } 
  }
  lappend tvlst0 [expr $chtvb];
  # dann spielen  
  for { set j 0; set k 0; set next [string index $chbt 0] }\
      { $next != "" } { incr j; set next [string index $chbt $j] } {
   if { $next == ":" } {
    incr k;
   } elseif { $next == "f" } {
    incr pifo;
   } elseif { $next == "p" } {
    incr pifo -1;
   } elseif { $next == "." } {
    if { $bthk != "" } {
     $bthk $rCTX $iocn; # beat hook proc
    }
    set tv0 [lindex $tvlst0 $k];
    incr k;
    set chtv0 [expr [lindex $tvlst0 $k]-$tv0];
    foreach {key dlc dur pif} $kcl {
     set tv1 [expr $tv0+$dlc*$chtv0];
     set pos [expr $tv1/$chtvb];
     set vl1 [off_volume $rCTX [expr $pifo+$pif] $pos]; 
     set ky1 [off_key $rCTX $key $pos]; 
     set tvdr [expr $dur*$chtv0];
     # Umrechnung von Notenzeit in Realzeit
     clock_advance $rCTX $tv1;
     set wh1 [clock_get_time $rCTX];
     set tv1 [clock_get_tval $rCTX];
     set dura [clock_advance $rCTX $tvdr]; 
     clock_set_time $rCTX $crtm; # Realzeit zuruecksetzen
     clock_set_tval $rCTX $crtv; # Notenzeit zuruecksetzen
     # Rueckwaertsspiel beruecksichtigen (negative Dauer)
     if { $dura < 0 } {
      set wh1 [expr $wh1+$dura];
      set dura [expr -$dura];
     }
     # Note spielen und Statistikdaten notieren
     instr_play_note $rCTX $iocn $inum $tv1 $tvdr $wh1\
            $dura $vl1 $loud $ky1 {} {};
     register_statistics $rCTX $dura $vl1 $ky1;
    }
   }
  }
  incr iocn; # Oktavierung hochzaehlen
 }
 instr_end_notes $rCTX;
 clock_advance $rCTX $chtv;
 return;
}

proc play_score score {
# spielt Noten
 upvar 1 rCTX rCTX;
 global $rCTX;
 # Rhythmus-Zweig bereitstellen
 set bell [TREEget $rCTX bell];
 if { $bell == "" } {
  # entweder einfaches Akkordspiel
  REF bell;
  array set $bell "STEM $rCTX song play_chord";
 } else {
  # oder vorgegebener Zweig
  global $bell;
  array set $bell "STEM $rCTX";
 }
 array set $rCTX "bell $bell";
 # Notenwert zuruecksetzen und Noten spielen
 TREEset $rCTX crtv 0; 
 source "notes/$score.tcl";
}

proc play args {
# allgemeine Spielprozedur
 # --- Initialisierungen ---
 upvar rCTX rCTX;
 if { [info exists rCTX] } {
  global $rCTX;
 }
 set out {};   # Ausgabekanal SCO-Datei
 set ctxnew 0; # Anzeige fuer neu angelegten Spielkontext

 # --- Schleife ueber alle Zweige ---
 foreach twig $args {
  # --- wenn noch kein Spielkontext existiert ... ---
  if { ![info exists rCTX] || ![info exists $rCTX] } {
   # --- ... neuen Spielkontext anlegen mit
   #         Vorbelegung fuer alle notwendigen Eintraegen ---
   set rCTX [plant];
   set ctxnew 1;
  }
  # --- Spielkontext als Stamm merken ---
  set rStem $rCTX;
  # --- den Zweig in ein Feld umwandeln ---
  array set twig_array [concat $twig];
  # --- eventuelle weitere Abzweigung abschneiden und festhalten ---
  if { [array names twig_array more] != "" } {
   set more_args $twig_array(more); 
   array unset twig_array more;
  }
  # --- das Feld als neuen Zweig am Stamm befestigen ---
  set rCTX [TREEnew $rStem [array get twig_array]];
  # --- wenn noch nicht geschehen, jetzt die SCO-Datei oeffnen ---
  if { [TREEget $rCTX out] == {} } {
   set out [open_sco_file $rCTX];
  }
  # --- falls weitere Abzweigung vorhanden, ... ---
  if { [info exists more_args] } {
   # --- ... das Spiel dorthin abgeben ... ---
   puts "+ Verzweigung beginnt"
   eval "play $more_args";
   puts "- Verzweigung beendet"
  } else {
   # --- ... sonst aktuelle Vorspiel- Lied- und Nachspiel-Prozedur spielen ---
   # --- Vorspiel --- 
   set prld [TREEget $rCTX prld];
   if { $prld != "" } {
    puts "+ Vorspiel $prld : [clock_set_stamp $rCTX] s";
    eval $prld;
    puts "- Vorspiel $prld : [clock_get_time $rCTX] s";
   } else {
    puts "*** kein Vorspiel ***";
   }
   # --- Lied ---
   set song [TREEget $rCTX song];
   if { $song != "" } {
    puts "+ Lied $song : [clock_set_stamp $rCTX] s";
    eval $song;
    puts "- Lied $song : [clock_get_time $rCTX] s";
   } else {
    puts "*** kein Lied ***";
   }
   # --- Nachspiel ---
   set psld [TREEget $rCTX psld];
   if { $psld != "" } {
    puts "+ Nachspiel $psld : [clock_set_stamp $rCTX] s";
    eval $psld;
    puts "- Nachspiel $psld : [clock_get_time $rCTX] s";
   } else {
    puts "*** kein Nachspiel ***";
   }
  }
  # --- wenn die SCO-Datei in diesem Lauf geoeffnet wurde ... ---
  if { $out != {} } {
   # --- ... jetzt ORC-Datei schreiben und SCO-Datei wieder schliessen ---
   write_orc_file $rCTX;
   set out [close_sco_file $rCTX];
  }
  # --- den neuen Zweig wieder abschneiden und verwerfen ---
  TREEerase $rCTX;
  set rCTX $rStem;
  # --- ebenso das Feld fuer den naechsten Zweig ---
  array unset twig_array;
 }
 # --- Gesamtspielzeit als Rueckgabewert merken
 set result [TREEget $rCTX tttm];
 # --- wenn in diesem Lauf eine neuer Kontext angelegt wurde ... ---
 if { $ctxnew } {
  # --- ... diesen nun wieder verwerfen ---
  TREEerase $rStem;
 }
 return $result;
}

# ----------------------------------------------------------------------------
puts "Lektion x : Klangbaeume, sound trees";
# ----------------------------------------------------------------------------

proc plant {} {
# legt einen neuen Spielkontext an (pflanzt einen neuen Klangbaum), 
# mit Vorbelegung aller notwendigen Eintraege
 REF rCTX;
 TREEnew $rCTX { 
  loca  music
  tttm  0
  out   {}
  instr {}
  crtm  0.0
  sttm  0.0
  crtv  0
  sttv  0
  lega  0
  fort  0
  velo  0
  opng  {play_opening}
  prld  {play_pre}
  song  {play_chord}
  psld  {play_post}
  clng  {play_closing}
  bthk  {}  
  inam  ethereal_organ
  tmsg  4/4
  tmtv  2  
  tvtv  1
  tmfn  time_function
  kybs  0
  kyfn  key_function
  btbs  0
  vlbs  76
  cres  0
  fstp  3
  vlfn  volume_function
  loud  1.0
  vstp  1.2
  frmt  3
  optv  0.0
  prtv  0.0
  pstv  0.0
  cltv  0.5
  pmod  melody
  csym  C
  chtv  1
  chbt  {}
  cmod  {}
  cstc  .4
  text  {}
  idst  100
  iazm  0
  isze  100
  iang  0
 }
 return $rCTX;
}

proc sprout {rCTX args} {
# fuegt einem Spielkontext weitere Zweige hinzu (laesst Zweige spriessen)
 global $rCTX;
 # --- Schleife ueber alle Zweige ---
 foreach twig $args {
  array set twig_array [concat $twig];         # Zweigliste in Feld wandeln
  if { [array names twig_array more] != "" } { # weitere Zweige
   set more_args $twig_array(more);            # am Zweig selbst
   array unset twig_array more;                # aus dem Feld entfernen
  }
  set rtwig [TREEnew $rCTX [array get twig_array]];  # diesen Zweig
  if { [info exists more_args] } {                   # und weitere Zweige am
   eval "sprout $rtwig $more_args";                  # Zweig selbst hinzufuegen
   unset more_args;                                  # und vergessen
  }
  array unset twig_array;                      # Feld ausleeren
 }
 return [set ${rCTX}(TWIGS)];                  # Zweigliste ausgeben
}

proc begin {rCTX loca optv} {
# bereitet das Spiel vor
 global $rCTX;
 TREEset $rCTX loca $loca;
 TREEset $rCTX optv $optv;
 # Orchester-Datei oeffnen
 open_sco_file $rCTX;
 # Eroeffnungsprozedur spielen
 set opng [TREEget $rCTX opng];
 if { $opng != "" } {
  puts "+ Eroeffnung : [clock_set_stamp $rCTX] s";
  eval $opng;
  puts "- Eroeffnung : [clock_get_time $rCTX] s";
 } else {
  puts "*** keine Eroeffnung ***";
 }
 return;
}

proc touch {rCTX num} {
# spielt einen Klangbaum $num mal (schuettelt den Baum)
 global $rCTX;
 # --- Variantenliste und Stamm bereitstellen ---
 if { [array names $rCTX vars] != "" } {
  set vars [set ${rCTX}(vars)];
 } else {
  set vars [list];
 }
 set nvars [llength $vars];
 if { $nvars > 0 } {
  if { [array names $rCTX STEM] == "" } {
   REF stem;
   array set $rCTX "STEM $stem";
  } else {
   set stem [set ${rCTX}(STEM)];
   global $stem;
  }
 }
 # --- Schuettelschleife ---
 for { set i 0 } { $i < $num } { incr i } {
  # --- Stamm mit naechster Variante ueberschreiben ---
  if { $nvars > 0 } {
   if { [array names $rCTX vind] == "" } {
    array set $rCTX "vind 0";   
   }
   set vind [set ${rCTX}(vind)]; 
   set var [lindex $vars $vind];
   array set $stem [concat $var];
   set vind [expr ($vind+1)%$nvars];
   array set $rCTX "vind $vind";
   puts [TREEget $rCTX out] "; variation : $var";
  }
  if { [array names $rCTX TWIGS] == "" } {
   # --- wenn es keine weiteren Zweige gibt,
   #     das Lied mit Vorspiel und Nachspiel spielen ---
   # --- Vorspiel --- 
   set prld [TREEget $rCTX prld];
   if { $prld != "" } {
    puts "+ Vorspiel $prld : [clock_set_stamp $rCTX] s";
    eval $prld;
    puts "- Vorspiel $prld : [clock_get_time $rCTX] s";
   } else {
    puts "*** kein Vorspiel ***";
   }
   # --- Lied ---
   set song [TREEget $rCTX song];
   if { $song != "" } {
    puts "+ Lied $song : [clock_set_stamp $rCTX] s";
    eval $song;
    puts "- Lied $song : [clock_get_time $rCTX] s";
   } else {
    puts "*** kein Lied ***";
   }
   # --- Nachspiel ---
   set psld [TREEget $rCTX psld];
   if { $psld != "" } {
    puts "+ Nachspiel $psld : [clock_set_stamp $rCTX] s";
    eval $psld;
    puts "- Nachspiel $psld : [clock_get_time $rCTX] s";
   } else {
    puts "*** kein Nachspiel ***";
   }
  } else {
   # --- sonst den Zweig schuetteln, der an der Reihe ist ---
   if { [array names $rCTX turn] == "" } {
    array set $rCTX {turn 0};
   }
   set turn  [set ${rCTX}(turn)];
   if { $turn >= 0 } {
    set twigs [set ${rCTX}(TWIGS)];
    set len   [llength $twigs];
    set turn  [expr $turn%$len];
    set rtwig [lindex $twigs $turn];
    global $rtwig;
    # --- Wiederholungszaehler bereitstellen ---
    if { [array names $rtwig rept] == "" } {
     array set $rtwig {rept 1}
    }     
    set rept [set ${rtwig}(rept)];
    if { [array names $rtwig rlft] == "" } {
     array set $rtwig "rlft $rept"; 
    }
    set rlft [set ${rtwig}(rlft)];
    if { $rlft > 0 } {
     touch $rtwig 1;
     incr rlft -1;
    }
    # --- Wiederholungszaehler aktualisieren ---
    if { $rlft <= 0 } {
     incr turn;
     array set $rCTX "turn $turn";
     array unset $rtwig rlft;
    } else {
     array set $rtwig "rlft $rlft";
    }
   }
  }
 }
 return;
}

proc end {rCTX cltv} {
# schliesst das Spiel ab
 global $rCTX
 # Abschlussprozedur spielen
 TREEset $rCTX cltv $cltv;
 set clng [TREEget $rCTX clng];
 if { $clng != "" } {
  puts "+ Abschluss : [clock_set_stamp $rCTX] s";
  eval $clng;
  puts "- Abschluss : [clock_get_time $rCTX] s";
 } else {
  puts "*** kein Abschluss ***";
 }
 # Csound-Dateien abschliessen
 write_orc_file $rCTX;
 close_sco_file $rCTX;
 return;
}
