# ----------------------------------------------------------------------------
# Tcl Erweiterungs-Skript                                          Datenbaeume
# Copyright (c) 2002 by Hartmut Schwarz               matut@matut-spielwerk.de
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2 
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# ----------------------------------------------------------------------------

# *** setzt das Skript "ref.tcl" voraus (Referenzvariablen) ***

proc TREEnew { rstem leafs } {
# erzeugt einen neuen Zweig an einem existierenden Baum oder einen neuen Baum
# rstem = Referenz auf den Stamm (s. ref.tcl)
# leafs = Feldliste der Elemente (Blaetter) fuer den neuen Zweig oder Baum
 global $rstem;
 if { [array exists $rstem] } {
 # --- wenn es den Stamm gibt, einen neuen Zweig anbauen ---
   REF rtwig;
   global $rtwig;
   uplevel 1 "global $rtwig";
   lappend leafs STEM $rstem;
   array set $rtwig $leafs;        # neuer Zweig
   lappend ${rstem}(TWIGS) $rtwig; # am alten Stamm
   set result $rtwig;              # neue Referenz als Ergebnis
 } else {
 # --- wenn es den Stamm nicht gibt, einen neuen Baum erzeugen ---
  lappend leafs TWIGS [list];
  array set $rstem $leafs;
  set result $rstem;               # alte Referenz als Ergebnis
 }
 return $result;
}
proc TREEcut { rtwig } {
# schneidet einen Zweig von einem Baum ab und gibt ihn als Wurzel zurueck
 global $rtwig;
 if { [array names $rtwig STEM] != "" } {
  set rStem [set ${rtwig}(STEM)];
  global $rStem;
  if { [array names $rStem TWIGS] != "" } {
   set twigs [set ${rStem}(TWIGS)];   
   set i [lsearch -exact $twigs $rtwig];
   if { $i >= 0 } {
    set twigs [lreplace $twigs $i $i];
    set ${rStem}(TWIGS) $twigs; 
   }  
  }
  array unset $rtwig STEM;
 }
 return $rtwig;
}

proc TREEerase { rtwig } {
# schneidet einen Zweig von einem Baum ab und loescht ihn vollstaendig
 global $rtwig;
 TREEcut $rtwig;
 if { [array names $rtwig TWIGS] != "" } {
  foreach rtw [set ${rtwig}(TWIGS)] {
   TREEerase $rtw;
  }
 }
 array unset $rtwig "*";
 unset $rtwig;
 return $rtwig;
}

proc TREEget { rtwig leaf } {
# sucht in einen Baum von einem Zweig beginnend herab zur Wurzel nach
# einem bestimmten Blatt und liefert den Inhalt oder ""
# rtwig = Referenz auf den Zweig
# leaf  = Name des Blattes
 global $rtwig;
#puts "TREEget: rtwig = $rtwig, leaf = $leaf";
 # --- wenn sich das Blatt am Zweig gewendet hat ... ---
 if { [array names $rtwig $leaf] != "" } {
  # --- ... Wendeprozedur holen ---
  set cmd "set result \"[set ${rtwig}($leaf)]\""; # Wendeprozedur
#puts "Wendeprozedur: cmd = $cmd, Blaetter vom Stamm:";
  # --- Bis auf TWIGS und STEM alle Blaetter vom Stamm verfuegbar machen ---
  # H.S. OCT-21-2002 : zu aufwendig und auch unzureichend, weil so nicht auf
  # Variablen von unterhalb des Stammes zugegriffen werden kann. Nur der
  # Stamm selbst bleibt verfuegbar.
  if { [array names $rtwig STEM] != "" } {
   set rstem [set ${rtwig}(STEM)];
   global $rstem;
  # foreach lf [array names $rstem] {
  #  if { ![string match STEM $lf] && ![string match TWIGS $lf] } {
  #   set $lf [TREEget $rstem $lf];
#puts "level = [info level], rstem = $rstem, lf = $lf";
  #  }
  # }
  }
  # --- damit dann die Wendeprozedur ausfuehren ---
  # H.S. OCT-21-2002 : ohne verfuegbare Stammvariablen ausfuehren, auf diese
  # kann aber ueber [STEM name] zugegriffen werden.
  eval $cmd;
 } elseif { [array names $rtwig STEM] != "" } {
  # --- ... sonst, falls moeglich, Blatt vom Stamm holen --- 
  set rstem [set ${rtwig}(STEM)];
#puts "Blatt vom Stamm holen: rstem = $rstem";
  global $rstem;
  set result [TREEget $rstem $leaf];
 } else {
  # --- ... oder, wenn gar nichts geht, ein leeres Blatt liefern ---
#puts "Nichts geht: Leerstring"
  set result "";
 }
#puts "Ergebnis: result = $result";

 # --- interner Zugriffsprozeduren auf
 #     Baumdaten innerhalb der Wendeprozedur ---
 proc THIS name {
  upvar rtwig rtwig;
  if { [info exists rtwig] } {
   return [TREEget $rtwig $name];
  }
  return "";
 }
 proc STEM name {
  upvar rstem rstem;
  if { [info exists rstem] } {
   return [TREEget $rstem $name];
  }
  return "";
 }
 # --- Schluss ---
 return $result;
}

proc TREEset { rtwig leaf value } {
# sucht in einem Baum von einem Zweig beginnend herab zur Wurzel nach einem
# bestimmten Blatt und aendert den Inhalt dieses Blattes auf den Inhalt von
# value. Wird das Blatt nicht gefunden, so wird ein neues Blatt an der Wurzel
# erzeugt.
 global $rtwig;
 if {   [array names $rtwig $leaf] != ""
     || [array names $rtwig STEM]  == "" } {
  set ${rtwig}($leaf) $value; 
 } else {
  TREEset [set ${rtwig}(STEM)] $leaf $value;
 }
 return $value;
}

proc TREEforeach { name tree body } {
# durch alle Zweige rauschen
 upvar $name nm;
 global $tree;
 # --- zuerst den Stamm hinauf ---
 set nm $tree;
 uplevel 1 $body;
 # --- dann durch die Zweige ---
 if { [array names $tree TWIGS] != "" } {
  foreach twig [set ${tree}(TWIGS)] {
   uplevel 1 "TREEforeach $name $twig {$body}";  
  }
 }
}

return;
# zum testen erst ref.tcl laden!!!
REF rBaum;
TREEnew $rBaum {a 1 b 2 c 3 d 4 e 5 };
set rZweig  [TREEnew $rBaum { d 40 e 50 f 60 }];
set rZweig2 [TREEnew $rZweig { a 10 b 20 }];  
