; --------------------------------------------------------------------------------
; Copyright (c) 2002 by Hartmut Schwarz               matut@matut-spielwerk.de
;
; This program is free software; you can redistribute it and/or
; modify it under the terms of the GNU General Public License
; as published by the Free Software Foundation; either version 2 
; of the License, or (at your option) any later version.
;
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
; GNU General Public License for more details.
;
; You should have received a copy of the GNU General Public
; License along with this program; if not, write to the Free
; Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
; MA 02111-1307, USA.
; --------------------------------------------------------------------------------

sr     = 44100
kr     = 11025
ksmps  =     4
nchnls =     2

; --------------------------------------------------------------------------------
         instr 1 ; Ein Klavier, ein Klavier !
                 ; 88 Tasten zwischen A2 (27.5 Hz) bis c5 (4186 Hz)
; --------------------------------------------------------------------------------
         ; eingelesene Parameter
ifree    = p3             ; Freischwingzeit ( p3 wird neu berechnet -> s.u. )
ivol     = p4             ; Lautstaerke
ipch     = p5             ; Tonhoehe
imix     = p6             ; Mixfaktor
idst     = p7             ; Hoerabstand
iazm     = p8             ; Richtungswinkel
isze     = p9             ; Groessenfaktor
iang     = p10            ; Aufstellungswinkel
         ; gebildete Parameter
iamp     = ampdb(ivol)               ; Amplitude
icps     = cpspch(ipch)              ; Frequenz
iatt     = (ivol>63?(ivol>75?3:2):1) ; Anschlagstaerke 1, 2 oder 3
         ; unabhaengige Werte
ift1     = 1             ; f-Tab Nummer f. Sinus
ift7     = 7             ; f-Tab Nummer f. Anstiegs-Phase
icar     = 1             ; Traeger-Koeffizient
imod     = 1             ; Modulator-Koeffizient
icdif    = 0.001*icps;   ; Differenz-Frequenz bei mehreren Saiten
icps0    = icps          ; Frequenz 1-te Saite
icps1    = icps-icdif    ; Frequenz 2-te Saite
icps2    = icps+icdif    ; Frequenz 3-te Saite
iye      = .001          ; normierte Endamplitude
         ; frequenzabhaengige Werte
ibalance = .5-.00004*(2500-icps)     ; Kanal-Balance ( 0 = links ... 1 = rechts )
im27     = (iatt>1?(iatt>2?3:2):1)   ; Modulationsindex bei 27.5 Hz
imind    = 1+im27*(27.5/icps)        ; Modulationsindex
ia       = 1+16*(icps/4186)^2        ; Daempfungskoeffizient
ib       = ia                        ; Freischwing-Daempfung ( 0 = keine )
ic       = 100*ia                    ; Daempfer-Daempfung
iy0      = 1+((icps-2080)/4186)^2    ; normierte Anfangsamplitude
iy0      = .38*iy0                   ; Angleichung an my_piano !!! 
ir4186   = (iatt>1?(iatt>2?0.0005:0.001):0.002); Anstiegszeit bei 4186 Hz
irise    = ir4186*(4186/icps)        ; frequenzabhaengige Anstiegszeit
         ; Daempfungsverlauf
iy1      = iy0*exp(-ib*ifree)        ; normierte Freischwing-Restamplitude
idec     = -(log(iye)+ib*ifree)/ic   ; Daempfungszeit
idec     = (idec < 0 ? .001 : idec)  ; wenn iye schon in Freischwing-Phase
idur     = irise + ifree + idec      ; Spieldauer neu berechnen
p3       = idur                      ; Standardparameter ueberschreiben !
         ; resultierende Einhuellende
kenv     envlpx iy0, irise, idur, idec, ift7, iy1/iy0, iye/iy1
kamp     = imix*kenv*iamp  ; fuer Amplitude
kmind    = kenv*imind ; und Modulationsindex gleichermassen !
         ; Signale fuer drei Saiten erzeugen und je nach Amplitude
         ; 1, 2 oder alle drei Signale ausgeben.
a0 foscili kamp, icps0, icar, imod, kmind, ift1       ; FM Oszillator 0
a1 foscili kamp, icps1, icar, imod, kmind, ift1       ; FM Oszillator 1
a2 foscili kamp, icps2, icar, imod, kmind, ift1       ; FM Oszillator 2
         ; compose signals main, left, right
amain    = (iatt>1?(iatt>2?(a0+a1+a2)/3:(a0+a1)/2):a0)
aleft    = (1 - ibalance)*amain
aright   = ibalance*amain
         ; signal output
outs     aleft, aright
endin ; --------------------------------------------------------------------------
